import pandas as pd
from pandas.core.frame import DataFrame, Series, Index
from .dffuzzmerge import *
from contextlib import suppress as contextlib_suppress


def get_name_not_in_cols(columns1, columns2):
    idxtmp1 = "___IDXTMP{number2replace}___"
    idxtmp2 = "___IDXTMP{number2replace}___"
    startnumber_1 = 0
    startnumber_2 = 1
    newcol1 = idxtmp1.format(number2replace=startnumber_1)
    newcol2 = idxtmp2.format(number2replace=startnumber_2)
    while newcol1 in columns1:
        startnumber_1 += 2
        newcol1 = idxtmp1.format(number2replace=startnumber_1)
    while newcol2 in columns2:
        startnumber_2 += 2
        newcol2 = idxtmp2.format(number2replace=startnumber_2)
    return newcol1, newcol2


def fuzzmatch(
    df1,
    df2,
    left_on,
    right_on,
    result_column="aa_match",
    idx_column2="fuzzidx2",
    function="ab_map_damerau_levenshtein_distance_2ways",
):
    if not isinstance(df2, DataFrame):
        df2 = pd.DataFrame(df2, columns=[right_on])
    df11 = df1.copy()
    df22 = df2.copy()
    df1 = (
        df1.loc[(df1[left_on].str.strip() != "")]
        .dropna(subset=[left_on])
        .drop_duplicates(subset=[left_on])
        .reset_index(drop=True)
        .copy()
    )
    df2 = (
        df2.loc[(df2[right_on].str.strip() != "")]
        .dropna(subset=[right_on])
        .drop_duplicates(subset=[right_on])
        .reset_index(drop=True)
        .copy()
    )
    idxtmp1, idxtmp2 = get_name_not_in_cols(df1.columns, df2.columns)
    pysm = PyStringMatcher(
        df1[left_on],
        df2[right_on],
    )
    q = getattr(pysm, function)()
    maplist1 = []
    maplist2 = []
    matchlist = []
    # print(df1)
    # print(df2)
    for x in q.values():
        with contextlib_suppress(Exception):
            realstring1 = df1.loc[x["aa_index_1"], left_on]
            realstring2 = df2.loc[x["aa_index_2"], right_on]
            result1 = df11.loc[df11[left_on] == realstring1].index
            result2 = df22.loc[df22[right_on] == realstring2].index[0]
            maplist1.extend(result1)
            maplist2.extend([result2 for _ in range(len(result1))])
            matchlist.extend((x["aa_match"] for _ in range(len(result1))))

    finalresult1 = (
        df11.loc[maplist1].assign(**{idxtmp1: maplist1}).reset_index(drop=True)
    )
    finalresult2 = (
        df22.loc[maplist2]
        .assign(**{idxtmp2: maplist2, result_column: matchlist})
        .reset_index(drop=True)
    )
    dffinal = pd.concat([finalresult1, finalresult2], axis=1)
    dffinal.index = dffinal[idxtmp1].__array__().copy()
    dffinal.drop(idxtmp1, axis=1, inplace=True)
    notfound = sorted(set(df11.index) - set(maplist1))
    dffinal2 = pd.concat([dffinal, df11.loc[notfound]], axis=0).loc[df11.index]
    dffinal2[idxtmp2] = dffinal2[idxtmp2].astype("Int64")
    return dffinal2.rename(columns={idxtmp2: idx_column2})


def d_fm_damerau_levenshtein_distance_1way(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="damerau_levenshtein_distance_1way_match",
        idx_column2="damerau_levenshtein_distance_1way_idx",
        function="ab_map_damerau_levenshtein_distance_1way",
    )


def d_fm_damerau_levenshtein_distance_2ways(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="damerau_levenshtein_distance_2ways_match",
        idx_column2="damerau_levenshtein_distance_2ways_idx",
        function="ab_map_damerau_levenshtein_distance_2ways",
    )


def d_fm_hemming_distance_1way(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="hemming_distance_1way_match",
        idx_column2="hemming_distance_1way_idx",
        function="ab_map_hemming_distance_1way",
    )


def d_fm_hemming_distance_2ways(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="hemming_distance_2ways_match",
        idx_column2="hemming_distance_2ways_idx",
        function="ab_map_hemming_distance_2ways",
    )


def d_fm_jaro_2ways(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="jaro_2ways_match",
        idx_column2="jaro_2ways_idx",
        function="ab_map_jaro_2ways",
    )


def d_fm_jaro_distance_1way(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="jaro_distance_1way_match",
        idx_column2="jaro_distance_1way_idx",
        function="ab_map_jaro_distance_1way",
    )


def d_fm_jaro_winkler_distance_1way(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="jaro_winkler_distance_1way_match",
        idx_column2="jaro_winkler_distance_1way_idx",
        function="ab_map_jaro_winkler_distance_1way",
    )


def d_fm_jaro_winkler_distance_2ways(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="jaro_winkler_distance_2ways_match",
        idx_column2="jaro_winkler_distance_2ways_idx",
        function="ab_map_jaro_winkler_distance_2ways",
    )


def d_fm_levenshtein_distance_1way(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="levenshtein_distance_1way_match",
        idx_column2="levenshtein_distance_1way_idx",
        function="ab_map_levenshtein_distance_1way",
    )


def d_fm_levenshtein_distance_2ways(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="levenshtein_distance_2ways_match",
        idx_column2="levenshtein_distance_2ways_idx",
        function="ab_map_levenshtein_distance_2ways",
    )


def d_fm_longest_common_subsequence_v0(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="longest_common_subsequence_v0_match",
        idx_column2="longest_common_subsequence_v0_idx",
        function="ab_map_longest_common_subsequence_v0",
    )


def d_fm_longest_common_substring_v0(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="longest_common_substring_v0_match",
        idx_column2="longest_common_substring_v0_idx",
        function="ab_map_longest_common_substring_v0",
    )


def d_fm_longest_common_substring_v1(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="longest_common_substring_v1_match",
        idx_column2="longest_common_substring_v1_idx",
        function="ab_map_longest_common_substring_v1",
    )


def d_fm_subsequence_v1(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="subsequence_v1_match",
        idx_column2="subsequence_v1_idx",
        function="ab_map_subsequence_v1",
    )


def d_fm_subsequence_v2(
    self,
    df2,
    left_on,
    right_on,
):
    return fuzzmatch(
        df1=self,
        df2=df2,
        left_on=left_on,
        right_on=right_on,
        result_column="subsequence_v2_match",
        idx_column2="subsequence_v2_idx",
        function="ab_map_subsequence_v2",
    )


DataFrame.d_fm_damerau_levenshtein_distance_1way = (
    d_fm_damerau_levenshtein_distance_1way
)
DataFrame.d_fm_damerau_levenshtein_distance_2ways = (
    d_fm_damerau_levenshtein_distance_2ways
)
DataFrame.d_fm_hemming_distance_1way = d_fm_hemming_distance_1way
DataFrame.d_fm_hemming_distance_2ways = d_fm_hemming_distance_2ways
DataFrame.d_fm_jaro_2ways = d_fm_jaro_2ways
DataFrame.d_fm_jaro_distance_1way = d_fm_jaro_distance_1way
DataFrame.d_fm_jaro_winkler_distance_1way = d_fm_jaro_winkler_distance_1way
DataFrame.d_fm_jaro_winkler_distance_2ways = d_fm_jaro_winkler_distance_2ways
DataFrame.d_fm_levenshtein_distance_1way = d_fm_levenshtein_distance_1way
DataFrame.d_fm_levenshtein_distance_2ways = d_fm_levenshtein_distance_2ways
DataFrame.d_fm_longest_common_subsequence_v0 = d_fm_longest_common_subsequence_v0
DataFrame.d_fm_longest_common_substring_v0 = d_fm_longest_common_substring_v0
DataFrame.d_fm_longest_common_substring_v1 = d_fm_longest_common_substring_v1
DataFrame.d_fm_subsequence_v1 = d_fm_subsequence_v1
DataFrame.d_fm_subsequence_v2 = d_fm_subsequence_v2



import os
import subprocess
import json
import requests

this_folder = os.path.dirname(os.path.abspath(__file__))
json_file = os.path.join(this_folder, "cpp_parsers.json")
link_uiautomator_dump_without_could_not_detect_idle_state = "https://github.com/hansalemaos/uiautomator_dump_without_could_not_detect_idle_state/raw/refs/heads/main/uiautomatorcpulimit.cpp"
link_uiautomator_dump_to_csv = "https://github.com/hansalemaos/uiautomator_dump_to_csv/raw/refs/heads/main/uiautomatornolimit.cpp"
link_sendevent_multicommands_type_text = "https://github.com/hansalemaos/sendevent_multicommands_type_text/raw/refs/heads/main/sendeventall.cpp"
link_mouse_sendevent_android = "https://github.com/hansalemaos/mouse_sendevent_android/raw/refs/heads/main/sendeventmouse.cpp"
link_getevent_keydumper_linux = "https://github.com/hansalemaos/getevent_keydumper_linux/raw/refs/heads/main/newgetevent.cpp"
link_getevent_pretty_print_linux = "https://github.com/hansalemaos/getevent_pretty_print_linux/raw/refs/heads/main/geteventall.cpp"
link_android_fragment_parser = "https://github.com/hansalemaos/android_fragment_parser/raw/refs/heads/main/fragmentdumper.cpp"
link_uiautomator_eventsparser_cpp = "https://github.com/hansalemaos/uiautomator_eventsparser_cpp/raw/refs/heads/main/eventparser.cpp"
link_lcp = "https://github.com/hansalemaos/elparse/raw/refs/heads/main/elparse.cpp"
link_lcp_apk = (
    "https://github.com/hansalemaos/elparse/raw/refs/heads/main/app-release.apk"
)
link_uiautomator2tocsv = "https://github.com/hansalemaos/uiautomator2tocsv/raw/refs/heads/main/uiautomator2parser.cpp"
link_screencaptoppm = "https://github.com/hansalemaos/screencap_uncompressed_screenshots/raw/refs/heads/main/screencaptoppm.cpp"
link_hocr2csv = "https://github.com/hansalemaos/screencap2tesseract2csv/raw/refs/heads/main/tesseractwithimagemagick.cpp"

folder_uiautomator_dump_without_could_not_detect_idle_state = os.path.join(
    this_folder, "uiautomator_dump_without_could_not_detect_idle_state"
)
folder_uiautomator_dump_to_csv = os.path.join(this_folder, "uiautomator_dump_to_csv")
folder_sendevent_multicommands_type_text = os.path.join(
    this_folder, "sendevent_multicommands_type_text"
)
folder_mouse_sendevent_android = os.path.join(this_folder, "mouse_sendevent_android")
folder_getevent_keydumper_linux = os.path.join(this_folder, "getevent_keydumper_linux")
folder_getevent_pretty_print_linux = os.path.join(
    this_folder, "getevent_pretty_print_linux"
)
folder_android_fragment_parser = os.path.join(this_folder, "android_fragment_parser")
folder_uiautomator_eventsparser_cpp = os.path.join(
    this_folder, "uiautomator_eventsparser_cpp"
)
folder_lcp = os.path.join(this_folder, "lcp")
folder_uiautomator2tocsv = os.path.join(this_folder, "uiautomator2tocsv")
folder_screencaptoppm = os.path.join(this_folder, "screencaptoppm")
folder_hocr2csv = os.path.join(this_folder, "hocr2csv")


def _download_and_save(link, folder, filename):
    filepath = os.path.join(
        folder,
        filename,
    )
    with requests.get(link) as r:
        with open(
            filepath,
            "wb",
        ) as f:
            f.write(r.content)
    return filepath


def download_uiautomator_dump_without_could_not_detect_idle_state():
    return _download_and_save(
        link_uiautomator_dump_without_could_not_detect_idle_state,
        folder_uiautomator_dump_without_could_not_detect_idle_state,
        "cppfile.cpp",
    )


def download_uiautomator_dump_to_csv():
    return _download_and_save(
        link_uiautomator_dump_to_csv,
        folder_uiautomator_dump_to_csv,
        "cppfile.cpp",
    )


def download_sendevent_multicommands_type_text():
    return _download_and_save(
        link_sendevent_multicommands_type_text,
        folder_sendevent_multicommands_type_text,
        "cppfile.cpp",
    )


def download_mouse_sendevent_android():
    return _download_and_save(
        link_mouse_sendevent_android,
        folder_mouse_sendevent_android,
        "cppfile.cpp",
    )


def download_getevent_pretty_print_linux():
    return _download_and_save(
        link_getevent_pretty_print_linux,
        folder_getevent_pretty_print_linux,
        "cppfile.cpp",
    )


def download_android_fragment_parser():
    return _download_and_save(
        link_android_fragment_parser,
        folder_android_fragment_parser,
        "cppfile.cpp",
    )


def download_uiautomator_eventsparser_cpp():
    return _download_and_save(
        link_uiautomator_eventsparser_cpp,
        folder_uiautomator_eventsparser_cpp,
        "cppfile.cpp",
    )


def download_lcp():
    return _download_and_save(
        link_lcp,
        folder_lcp,
        "cppfile.cpp",
    )


def download_getevent_keydumper_linux():
    return _download_and_save(
        link_getevent_keydumper_linux,
        folder_getevent_keydumper_linux,
        "cppfile.cpp",
    )


def download_uiautomator2tocsv():
    return _download_and_save(
        link_uiautomator2tocsv,
        folder_uiautomator2tocsv,
        "cppfile.cpp",
    )


def download_screencaptoppm():
    return _download_and_save(
        link_screencaptoppm,
        folder_screencaptoppm,
        "cppfile.cpp",
    )


def download_hocr2csv():
    return _download_and_save(
        link_hocr2csv,
        folder_hocr2csv,
        "cppfile.cpp",
    )


def download_all_scripts():
    return (
        download_uiautomator_dump_without_could_not_detect_idle_state(),
        download_uiautomator_dump_to_csv(),
        download_sendevent_multicommands_type_text(),
        download_mouse_sendevent_android(),
        download_getevent_keydumper_linux(),
        download_getevent_pretty_print_linux(),
        download_android_fragment_parser(),
        # download_uiautomator_eventsparser_cpp(),
        download_lcp(),
        download_uiautomator2tocsv(),
        # download_screencaptoppm(),
        download_hocr2csv(),
    )


def compile_files(
    allscripts,
    compiler=r"g++",
    compiler_args=(
        "-std=c++2a",
        "-O3",
        "-g0",
        "-march=native",
        "-mtune=native",
        "-o",
        "a.out",
    ),
):
    compiler_args = (compiler,) + compiler_args
    all_executable_files = {}
    current_working_dict = os.getcwd()
    for script in allscripts:
        next_working_dict = os.path.dirname(script)
        os.chdir(next_working_dict)
        cmd2execute = compiler_args + (script,)
        subprocess.run(
            " ".join(cmd2execute),
            shell=True,
            env=os.environ,
            cwd=next_working_dict,
        )
        executable_file = os.path.join(next_working_dict, "a.out")
        spli = next_working_dict.split(os.sep)
        all_executable_files[spli[len(spli) - 1]] = executable_file
    os.chdir(current_working_dict)
    return all_executable_files


def download_and_compile_files(gcc_exe="g++"):
    allscripts = download_all_scripts()
    jsondata = compile_files(
        allscripts,
        compiler=gcc_exe,
        compiler_args=(
            "-std=c++2a",
            "-O3",
            "-g0",
            "-march=native",
            "-mtune=native",
            "-o",
            "a.out",
        ),
    )

    with open(json_file, mode="w", encoding="utf-8") as f:
        f.write(json.dumps(jsondata))
    return jsondata


def makedirs_for_plugins():
    os.makedirs(folder_screencaptoppm, exist_ok=True)
    os.makedirs(folder_hocr2csv, exist_ok=True)
    os.makedirs(
        folder_uiautomator_dump_without_could_not_detect_idle_state, exist_ok=True
    )
    os.makedirs(folder_uiautomator_dump_to_csv, exist_ok=True)
    os.makedirs(folder_sendevent_multicommands_type_text, exist_ok=True)
    os.makedirs(folder_mouse_sendevent_android, exist_ok=True)
    os.makedirs(folder_getevent_keydumper_linux, exist_ok=True)
    os.makedirs(folder_getevent_pretty_print_linux, exist_ok=True)
    os.makedirs(folder_android_fragment_parser, exist_ok=True)
    os.makedirs(folder_uiautomator_eventsparser_cpp, exist_ok=True)
    os.makedirs(folder_lcp, exist_ok=True)
    os.makedirs(folder_getevent_keydumper_linux, exist_ok=True)
    os.makedirs(folder_uiautomator2tocsv, exist_ok=True)


def load_parser_paths():
    with open(json_file, "r", encoding="utf-8") as f:
        data = f.read()
    return json.loads(data)


def load_parser_paths_or_compile(gcc_exe="g++"):
    makedirs_for_plugins()
    try:
        return load_parser_paths()
    except Exception:
        download_and_compile_files(gcc_exe=gcc_exe)
        return load_parser_paths()



import sys
import traceback
import os


class JustColors:
    def __init__(self):
        self.BOLD = "\033[1m"
        self.ITALIC = "\033[3m"
        self.UNDERLINE = "\033[4m"
        self.UNDERLINE_THICK = "\033[21m"
        self.HIGHLIGHTED = "\033[7m"
        self.HIGHLIGHTED_BLACK = "\033[40m"
        self.HIGHLIGHTED_RED = "\033[41m"
        self.HIGHLIGHTED_GREEN = "\033[42m"
        self.HIGHLIGHTED_YELLOW = "\033[43m"
        self.HIGHLIGHTED_BLUE = "\033[44m"
        self.HIGHLIGHTED_PURPLE = "\033[45m"
        self.HIGHLIGHTED_CYAN = "\033[46m"
        self.HIGHLIGHTED_GREY = "\033[47m"
        self.HIGHLIGHTED_GREY_LIGHT = "\033[100m"
        self.HIGHLIGHTED_RED_LIGHT = "\033[101m"
        self.HIGHLIGHTED_GREEN_LIGHT = "\033[102m"
        self.HIGHLIGHTED_YELLOW_LIGHT = "\033[103m"
        self.HIGHLIGHTED_BLUE_LIGHT = "\033[104m"
        self.HIGHLIGHTED_PURPLE_LIGHT = "\033[105m"
        self.HIGHLIGHTED_CYAN_LIGHT = "\033[106m"
        self.HIGHLIGHTED_WHITE_LIGHT = "\033[107m"
        self.STRIKE_THROUGH = "\033[9m"
        self.MARGIN_1 = "\033[51m"
        self.MARGIN_2 = "\033[52m"
        self.BLACK = "\033[30m"
        self.RED_DARK = "\033[31m"
        self.GREEN_DARK = "\033[32m"
        self.YELLOW_DARK = "\033[33m"
        self.BLUE_DARK = "\033[34m"
        self.PURPLE_DARK = "\033[35m"
        self.CYAN_DARK = "\033[36m"
        self.GREY_DARK = "\033[37m"
        self.BLACK_LIGHT = "\033[90m"
        self.RED = "\033[91m"
        self.GREEN = "\033[92m"
        self.YELLOW = "\033[93m"
        self.BLUE = "\033[94m"
        self.PURPLE = "\033[95m"
        self.CYAN = "\033[96m"
        self.WHITE = "\033[97m"
        self.DEFAULT = "\033[0m"


mycolors = JustColors()


def printincolor(values, color=None, print_to_stderr=False):
    s1 = "GOT AN ERROR DURING PRINTING"
    if color:
        try:
            s1 = "%s%s%s" % (color, values, mycolors.DEFAULT)
        except Exception:
            if isinstance(values, bytes):
                s1 = "%s%s%s" % (
                    color,
                    values.decode("utf-8", "backslashreplace"),
                    mycolors.DEFAULT,
                )
            else:
                s1 = "%s%s%s" % (color, repr(values), mycolors.DEFAULT)
        if print_to_stderr:
            sys.stderr.flush()
            sys.stderr.write(f"{s1}\n")
            sys.stderr.flush()
        else:
            print(s1)

    else:
        try:
            s1 = "%s%s" % (values, mycolors.DEFAULT)
        except Exception:
            if isinstance(values, bytes):
                s1 = "%s%s" % (
                    values.decode("utf-8", "backslashreplace"),
                    mycolors.DEFAULT,
                )
            else:
                s1 = "%s%s" % (repr(values), mycolors.DEFAULT)
        if print_to_stderr:
            sys.stderr.flush()
            sys.stderr.write(f"{s1}\n")
            sys.stderr.flush()
        else:
            print(s1)


def errwrite(*args, **kwargs):
    symbol_top = kwargs.pop("symbol_top", "╦")
    symbol_bottom = kwargs.pop("symbol_bottom", "╩")
    len_top = kwargs.pop("len_top", "60")
    len_bottom = kwargs.pop("len_bottom", "60")
    color_top = kwargs.pop("color_top", "YELLOW_DARK")
    color_bottom = kwargs.pop("color_bottom", "RED_DARK")
    print_to_stderr = kwargs.pop("print_to_stderr", False)
    color_exception = kwargs.pop("color_exception", "CYAN")
    color2print_top = None
    color2print_bottom = None
    color_exceptionmiddle = None
    try:
        color2print_top = mycolors.__dict__.get(
            color_top, mycolors.__dict__.get("YELLOW_DARK")
        )
        color2print_bottom = mycolors.__dict__.get(
            color_bottom, mycolors.__dict__.get("RED_DARK")
        )
        color_exceptionmiddle = mycolors.__dict__.get(
            color_exception, mycolors.__dict__.get("CYAN")
        )
    except Exception as e:
        print(e)

    printincolor(
        values="".join(symbol_top * int(len_top)),
        color=color2print_top,
        print_to_stderr=print_to_stderr,
    )
    etype, value, tb = sys.exc_info()
    lines = traceback.format_exception(etype, value, tb)
    try:
        if print_to_stderr:
            sys.stderr.flush()

            sys.stderr.write("".join(lines))
            sys.stderr.flush()
        else:
            printincolor(
                "".join(lines),
                color=color_exceptionmiddle,
                print_to_stderr=print_to_stderr,
            )
    except Exception:
        print("".join(lines))
    printincolor(
        "".join(symbol_bottom * int(len_bottom)),
        color=color2print_bottom,
        print_to_stderr=print_to_stderr,
    )



import threading, sys, ctypes


def kill_thread(threadobject):
    # based on https://pypi.org/project/kthread/
    if not threadobject.is_alive():
        return True
    tid = -1
    for tid1, tobj in threading._active.items():
        if tobj is threadobject:
            tid = tid1
            break
    if tid == -1:
        sys.stderr.write(f"{threadobject} not found")
        return False
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(tid), ctypes.py_object(SystemExit)
    )
    if res == 0:
        return False
    elif res != 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        return False
    return True



cimport cython
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map
from libc.stdint cimport int64_t
from libcpp.utility cimport pair
import os.path
# https://www.youtube.com/watch?v=sWgDk-o-6ZE  (21:20)
cdef extern from "fuzzmatcher.h" nogil :
    cdef cppclass StringMatcher:
        StringMatcher(vector[string]&, vector[string]&)
        void _load_vecs_for_cython(vector[string]*, vector[string]*)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_longest_common_substring_v1(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_longest_common_substring_v1(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_hemming_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_hemming_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_hemming_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_hemming_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_longest_common_substring_v0(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_longest_common_substring_v0(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_longest_common_subsequence_v0(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_longest_common_subsequence_v0(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_damerau_levenshtein_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_damerau_levenshtein_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_damerau_levenshtein_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_damerau_levenshtein_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_levenshtein_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_levenshtein_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_levenshtein_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_levenshtein_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_jaro_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_jaro_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_jaro_winkler_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_jaro_winkler_distance_1way(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_jaro_winkler_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_jaro_winkler_distance_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_jaro_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_jaro_2ways(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_subsequence_v1(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_subsequence_v1(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ab_map_subsequence_v2(bint print_results, bint convert_to_csv, string file_path)
        unordered_map[int64_t, unordered_map[int64_t, pair[double, int64_t]]] ba_map_subsequence_v2(bint print_results, bint convert_to_csv, string file_path)
        StringMatcher& to_upper()
        StringMatcher& to_lower()
        StringMatcher& to_without_non_alphanumeric()
        StringMatcher& to_without_non_printable()
        StringMatcher& to_100_percent_copy()
        StringMatcher& to_without_whitespaces()
        StringMatcher& to_with_normalized_whitespaces()
        void _str__for_cython();


cdef extern from "fuzzmatcher.h" namespace "stringhelpers" nogil :
    vector[string] read_file_to_vector_lines(const string& filename)
    void _repr__for_cython(vector[string]*v);


cdef string convert_python_object_to_cpp_string(object shell_command):
    cdef:
        string cpp_shell_command
        bytes tmp_bytes
    if isinstance(shell_command,bytes):
        cpp_shell_command=<string>shell_command
    elif isinstance(shell_command,str):
        tmp_bytes=shell_command.encode()
        cpp_shell_command=<string>(tmp_bytes)
    else:
        tmp_bytes=str(shell_command).encode()
        cpp_shell_command=<string>(tmp_bytes)
    return cpp_shell_command



cdef void convert_to_stdvec(object stri, vector[string]& outvector):
    cdef:
        Py_ssize_t i
        string converted_cpp
    if isinstance(stri, (str, bytes)):
        stri = [stri]
    outvector.resize(len(stri))
    outvector.clear()
    for i in range(len(stri)):
        converted_cpp=convert_python_object_to_cpp_string(stri[i])
        outvector.emplace_back(converted_cpp)

cdef:
    string emptystring = <string>b""
    vector[string] emptystringvec = [b""]

@cython.final
cdef class PyStringMatcher:
    cdef:
        StringMatcher*sm
        vector[string] stri1list
        vector[string] stri2list

    def __cinit__(self):
        self.sm = new StringMatcher(emptystringvec,emptystringvec)
        self.stri1list=[]
        self.stri2list=[]

    def __init__(self, object stri1, object stri2):
        convert_to_stdvec(stri1,self.stri1list)
        convert_to_stdvec(stri2,self.stri2list)
        self.sm._load_vecs_for_cython(<vector[string]*>(&self.stri1list),<vector[string]*>(&self.stri2list))

    def __dealloc__(self):
        del self.sm

    def _filter_results(self, dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r, bint ab = True, bint sort_reverse=False):
        cdef:
            dict[str,object] outdict={}
            list sorteddict
            Py_ssize_t index
            bytes pystri1, pystri2
        sorteddict=sorted([[tuple(x[1].values())[0][0],x] for x in r.items()], key=lambda y: y[0],reverse=sort_reverse)
        for index in range(len(sorteddict)):
            if ab:
                pystri1=(self.stri1list[0][sorteddict[index][1][0]])
                pystri2=(self.stri2list[0][tuple(sorteddict[index][1][1].keys())[0]])
                outdict[index]={
                    "aa_match":sorteddict[index][0],
                    "aa_1_is_sub":tuple(sorteddict[index][1][1].values())[0][1],
                    "aa_index_1":sorteddict[index][1][0],
                    "aa_index_2":tuple(sorteddict[index][1][1].keys())[0],
                    "aa_str_1":pystri1.decode('utf-8','backslashreplace'),
                    "aa_str_2":pystri2.decode('utf-8','backslashreplace'),
                }
            else:
                pystri1=(self.stri2list[0][sorteddict[index][1][0]])
                pystri2=(self.stri1list[0][tuple(sorteddict[index][1][1].keys())[0]])
                outdict[index]={
                    "aa_match":sorteddict[index][0],
                    "aa_2_is_sub":tuple(sorteddict[index][1][1].values())[0][1],
                    "aa_index_2":sorteddict[index][1][0],
                    "aa_index_1":tuple(sorteddict[index][1][1].keys())[0],
                    "aa_str_2":pystri1.decode('utf-8','backslashreplace'),
                    "aa_str_1":pystri2.decode('utf-8','backslashreplace'),
                }
        return outdict

    def ab_map_damerau_levenshtein_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_damerau_levenshtein_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=False)

    def ab_map_damerau_levenshtein_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_damerau_levenshtein_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=False)

    def ab_map_hemming_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_hemming_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=False)

    def ab_map_hemming_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_hemming_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=False)

    def ab_map_jaro_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_jaro_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_jaro_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_jaro_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_jaro_winkler_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_jaro_winkler_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_jaro_winkler_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_jaro_winkler_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_levenshtein_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_levenshtein_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=False)

    def ab_map_levenshtein_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_levenshtein_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=False)

    def ab_map_longest_common_subsequence_v0(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_longest_common_subsequence_v0(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_longest_common_substring_v0(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_longest_common_substring_v0(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_longest_common_substring_v1(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_longest_common_substring_v1(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_subsequence_v1(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_subsequence_v1(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ab_map_subsequence_v2(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ab_map_subsequence_v2(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=True,sort_reverse=True)

    def ba_map_damerau_levenshtein_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_damerau_levenshtein_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=False)

    def ba_map_damerau_levenshtein_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_damerau_levenshtein_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=False)

    def ba_map_hemming_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_hemming_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=False)

    def ba_map_hemming_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_hemming_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=False)

    def ba_map_jaro_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_jaro_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_jaro_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_jaro_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_jaro_winkler_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_jaro_winkler_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_jaro_winkler_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_jaro_winkler_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_levenshtein_distance_1way(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_levenshtein_distance_1way(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=False)

    def ba_map_levenshtein_distance_2ways(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_levenshtein_distance_2ways(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=False)

    def ba_map_longest_common_subsequence_v0(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_longest_common_subsequence_v0(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_longest_common_substring_v0(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_longest_common_substring_v0(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_longest_common_substring_v1(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_longest_common_substring_v1(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_subsequence_v1(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_subsequence_v1(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def ba_map_subsequence_v2(self, bint print_cpp=False):
        cdef:
            dict[int64_t,dict[int64_t,tuple[double,int64_t]]] r
        r =self.sm.ba_map_subsequence_v2(print_cpp,False,emptystring)
        return self._filter_results(r=r,ab=False,sort_reverse=True)

    def cpp_data_to_upper(self):
        self.sm.to_upper()
        return self

    def cpp_data_to_lower(self):
        self.sm.to_lower()
        return self

    def cpp_data_to_without_non_alphanumeric(self):
        self.sm.to_without_non_alphanumeric()
        return self

    def cpp_data_to_without_non_printable(self):
        self.sm.to_without_non_printable()
        return self

    def cpp_data_to_100_percent_copy(self):
        self.sm.to_100_percent_copy()
        return self

    def cpp_data_to_without_whitespaces(self):
        self.sm.to_without_whitespaces()
        return self

    def cpp_data_to_with_normalized_whitespaces(self):
        self.sm.to_with_normalized_whitespaces()
        return self



